#pragma once

#include <functional>
#include <list>
#include <type_traits>
#include <vector>

#include <dune/common/hybridutilities.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>

#include <amdis/Boundary.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/common/ValueCategory.hpp>
#include <amdis/utility/RangeType.hpp>
#include <amdis/utility/TreeData.hpp>
#include <amdis/utility/Visitor.hpp>

namespace AMDiS
{
  /// Implements a boundary condition of Dirichlet-type.
  /**
   * By calling the methods \ref init() and \ref finish before and after
   * assembling the system-matrix, respectively, dirichlet boundary conditions
   * can be applied to the matrix and system vector. Therefore, a predicate
   * functions indicates the DOFs where values should be enforced and a second
   * functor provided in the constructor is responsible for determining the
   * values to be set at the DOFs.
   *
   * In the \ref finish method the matrix is called with \ref applyDirichletBC
   * to erase the corresponding rows and columns for the DOF indices. This
   * application of boundary conditions can be symmetric if the matrix does
   * support this symmetric modification. As a result, this method returns a list
   * of columns values, that should be subtracted from the rhs.
   **/
  template <class Domain, class Range = double>
  class DirichletBC
  {
  public:
    template <class Predicate, class Values,
      REQUIRES(Concepts::Functor<Predicate, bool(Domain)> &&
				       Concepts::Functor<Values,  Range(Domain)>) >
    DirichletBC(Predicate&& predicate, Values&& values)
      : predicate_(std::forward<Predicate>(predicate))
      , values_(std::forward<Values>(values))
    {}


    /// Prepare the matrix, rhs and solution block for assembling dirichlet
    /// boundary conditions, e.g. collect the corresponding indices of boundary
    /// DOFS for later elimination in \ref finish.
    template <class Matrix, class VectorX, class VectorB, class RowNode, class ColNode>
    void init(Matrix& /*matrix*/, VectorX& /*solution*/, VectorB& rhs,
              RowNode const& rowNode, ColNode const&)
    {
      if (!initialized_) {
        forEachLeafNode_(rowNode, [&](auto const& /*node*/, auto const& tp) {
          auto subBasis = Dune::Functions::subspaceBasis(rhs.basis(), cat(rowNode.treePath(),tp));
          this->initImpl(subBasis);
        });
        initialized_ = true;
      }
    }


    /// \brief Apply dirichlet BC to matrix and vector, i.e., add a unit-row
    /// to the matrix and optionally delete the corresponding matrix-column.
    template <class Matrix, class VectorX, class VectorB, class RowNode, class ColNode>
    void finish(Matrix& matrix, VectorX& solution, VectorB& rhs,
                RowNode const& rowNode, ColNode const& colNode)
    {
      test_exit_dbg(initialized_, "Boundary condition not initialized!");
      auto columns = matrix.applyDirichletBC(dirichletNodes_);

      Dune::Hybrid::ifElse(std::is_same<RangeType_t<RowNode>, Range>{},
        [&](auto id) {
          auto subBasis = Dune::Functions::subspaceBasis(rhs.basis(), rowNode.treePath());
          Dune::Functions::interpolate(subBasis, rhs, values_, dirichletNodes_);
        });

      Dune::Hybrid::ifElse(std::is_same<RangeType_t<ColNode>, Range>{},
        [&](auto id) {
          auto subBasis = Dune::Functions::subspaceBasis(solution.basis(), colNode.treePath());
          Dune::Functions::interpolate(subBasis, solution, values_, dirichletNodes_);
        });

      // subtract columns of dirichlet nodes from rhs
      for (auto const& triplet : columns)
        rhs[triplet.row] -= triplet.value * solution[triplet.col];

      initialized_ = false;
    }

  protected:
    // fill \ref dirichletNodes_ with 1 or 0 whether DOF corresponds to the boundary or not.
    template <class SubBasis>
    void initImpl(SubBasis const& subBasis)
    {
      using Dune::Functions::forEachBoundaryDOF;
      using LV = typename SubBasis::LocalView;
      using I = typename SubBasis::GridView::Intersection;
      dirichletNodes_.resize(subBasis.dimension());
      forEachBoundaryDOF(subBasis, [&](int localIndex, LV const& localView, I const& intersection) {
        dirichletNodes_[localView.index(localIndex)] = predicate_(intersection.geometry().center());
      });
    }

  private:
    std::function<bool(Domain)> predicate_;
    std::function<Range(Domain)> values_;

    bool initialized_ = false;
    std::vector<char> dirichletNodes_;
  };


  template <class Basis>
  struct DirichletData
  {
    using WorldVector = typename Basis::GridView::template Codim<0>::Geometry::GlobalCoordinate;

    template <class RowNode, class ColNode>
    using type = std::list< std::shared_ptr<DirichletBC<WorldVector, RangeType_t<RowNode>>> >;
  };

  template <class RowBasis, class ColBasis>
  using Constraints = MatrixData<RowBasis, ColBasis, DirichletData<RowBasis>::template type>;

} // end namespace AMDiS
