#pragma once

#include "ProblemStatBase.hpp"
#include "ProblemTimeInterface.hpp"

namespace AMDiS
{
  // forward declarations
  class AdaptInfo;

  /**
   * \ingroup Problem
   *
   * \brief
   * Base class for \ref ProblemInstat.
   */
  class ProblemInstatBase
      : public ProblemTimeInterface
      , public ProblemStatBase   // NOTE: Why is this derived from ProblemStatBase
  {
  public:
    /// Constructor.
    ProblemInstatBase(std::string const& name,
                      ProblemStatBase* initialProb)
      : name_(name)
      , initialProblem_(initialProb ? initialProb : this)
    {}

    /// Destructor.
    virtual ~ProblemInstatBase() = default;

    /// Implementation of \ref ProblemTimeInterface::setTime().
    virtual void setTime(AdaptInfo& adaptInfo) override;

    void solve(AdaptInfo&) { /* do nothing */ }

    /// Implementation of \ref ProblemStatBase::solve().
    virtual void solve(AdaptInfo& adaptInfo, bool, bool) override
    {
      solve(adaptInfo);
    }

    /// Implementation of \ref ProblemStatBase::estimate().
    virtual void estimate(AdaptInfo&) override { /* do nothing */ }

    /// Implementation of \ref ProblemStatBase::buildAfterAdapt().
    virtual void buildAfterAdapt(AdaptInfo&, Flag, bool, bool) override { /* do nothing */ }

    /// Implementation of \ref ProblemStatBase::markElements().
    virtual Flag markElements(AdaptInfo&) override
    {
      return 0;
    }

    /// Implementation of \ref ProblemStatBase::refineMesh().
    virtual Flag adaptGrid(AdaptInfo&) override
    {
      return 0;
    }

    /// Implementation of \ref ProblemTimeInterface::closeTimestep().
    virtual void closeTimestep(AdaptInfo&) override { /* do nothing */ }

    /// Implementation of \ref ProblemStatBase::name().
    virtual std::string const& name() const override
    {
      return name_;
    }

    /// Implementation of \ref ProblemTimeInterface::solveInitialProblem().
    virtual void solveInitialProblem(AdaptInfo& adaptInfo) override;

    /// Return reference to current simulation time \ref time_ set in \ref setTime
    /// from `adaptInfo->getTime()`.
    double const& time() const
    {
      return time_;
    }

    /// Return reference to current simulation timestep \ref tau_ set in \ref setTime
    /// from `adaptInfo->getTimestep()`.
    double const& tau() const&
    {
      return tau_;
    }

    /// Return reference to current simulation 1.0/timestep \ref invTau_ set in
    /// \ref setTime from `1.0 / adaptInfo->getTimestep()`.
    double const& invTau() const
    {
      return invTau_;
    }

  protected:
    /// Name of the problem.
    std::string name_;

    ProblemStatBase* initialProblem_;

    /// Time
    double time_ = 0.0;

    /// Timestep
    double tau_ = 1.0;

    /// 1 / timestep
    double invTau_ = 1.0;
  };

} // end namespace AMDiS
